-- TODO Canonical Forms

import .ftable
open nat


namespace lnat
  abbreviation var := nat

  -- Types --
  inductive typ : Type :=
  | nat : typ
  | arr : typ -> typ -> typ

  namespace typ
    definition decidable_eq [instance] : forall (t1 t2 : typ), decidable (t1 = t2)
    | typ.nat typ.nat := decidable.inl rfl
    | typ.nat (typ.arr l r) := decidable.inr (fun H, typ.no_confusion H)
    | (typ.arr l r) typ.nat := decidable.inr (fun H, typ.no_confusion H)
    | (typ.arr t1l t1r) (typ.arr t2l t2r) :=
        match decidable_eq t1l t2l with
        | decidable.inr H := decidable.inr (fun Heq, typ.no_confusion Heq (fun Heql, fun Heqr, H Heql))
        | decidable.inl Heql :=
            match decidable_eq t1r t2r with
            | decidable.inr H := decidable.inr (fun Heq, typ.no_confusion Heq (fun Heql, fun Heqr, H Heqr))
            | decidable.inl Heqr := decidable.inl (eq.subst Heqr (eq.subst Heql rfl))
            end
        end
  end typ

  abbreviation ctx := table.table var typ

  -- Constants --
  inductive const : Type :=
  | z
  | s

  namespace const
    inductive has_typ : const -> typ -> Prop :=
    | z : has_typ const.z typ.nat
    | s : has_typ const.s (typ.arr typ.nat typ.nat)

    namespace has_typ
      theorem z.inv : forall {t}, has_typ const.z t -> t = typ.nat
      | z.inv has_typ.z := rfl

      theorem s.inv : forall {t}, has_typ const.s t -> t = typ.arr typ.nat typ.nat
      | s.inv has_typ.s := rfl
    end has_typ

    theorem nat.canon : forall {op}, has_typ op typ.nat -> op = const.z
    | nat.canon has_typ.z := rfl

    theorem arr.canon : forall {op t1 t2}, has_typ op (typ.arr t1 t2) -> op = const.s /\ t1 = typ.nat /\ t2 = typ.nat
    | arr.canon has_typ.s := and.intro rfl (and.intro rfl rfl)
  end const

  -- Expressions --
  inductive exp : Type :=
  | var : var -> exp
  | const : const -> exp
  | lam : var -> typ -> exp -> exp
  | ap : exp -> exp -> exp

  namespace exp
    inductive has_typ : ctx -> exp -> typ -> Prop :=
    | var : Pi {g x t}, table.contains g x t -> has_typ g (exp.var x) t
    | const : Pi {c t}, Pi g, const.has_typ c t -> has_typ g (exp.const c) t
    | lam : Pi {g g' x t t' e},
        table.is_unionr g (table.singleton x t) g' ->
        has_typ g' e t' ->
        has_typ g (exp.lam x t e) (typ.arr t t')
    | ap : Pi {g t t' e1 e2},
        has_typ g e1 (typ.arr t t') ->
        has_typ g e2 t ->
        has_typ g (exp.ap e1 e2) t'

    namespace has_typ
      theorem var.inv : forall {c x t}, has_typ c (exp.var x) t -> table.contains c x t
      | var.inv (has_typ.var H) := H

      theorem const.z.inv : forall {c t}, has_typ c (exp.const (const.z)) t -> t = typ.nat
      | const.z.inv (has_typ.const c Hconst) := const.has_typ.z.inv Hconst

      theorem const.s.inv : forall {c t}, has_typ c (exp.const (const.s)) t -> t = typ.arr typ.nat typ.nat
      | const.s.inv (has_typ.const c Hconst) := const.has_typ.s.inv Hconst

      theorem lam.inv : forall {c x e t t''}, has_typ c (exp.lam x t e) t'' -> exists c' t',
        (table.is_unionr c (table.singleton x t) c') /\
        (has_typ c' e t') /\
        (t'' = typ.arr t t')
      | lam.inv (@has_typ.lam c c' x t t' e H1 H2) := exists.intro c' (exists.intro t' (and.intro H1 (and.intro H2 rfl)))

      theorem ap.inv : forall {c t' e1 e2}, has_typ c (exp.ap e1 e2) t' -> exists t, (has_typ c e1 (typ.arr t t') /\ has_typ c e2 t)
      | ap.inv (@has_typ.ap g t t' e1 e2 H1 H2) := exists.intro t (and.intro H1 H2)

      -- Unicity of Typing
      theorem inj : forall {c e t t'}, has_typ c e t -> has_typ c e t' -> t = t'
      | inj (@has_typ.var c x t Hctx) (@has_typ.var c x t' Hctx') :=
          table.contains.inj Hctx Hctx'
      | inj (@has_typ.const (const.z) t c Hconst) (@has_typ.const (const.z) t' c Hconst') :=
          have P : t = typ.nat, from const.has_typ.z.inv Hconst,
          have P' : t' = typ.nat, from const.has_typ.z.inv Hconst',
          eq.subst (eq.symm P') P
      | inj (@has_typ.const (const.s) t c Hconst) (@has_typ.const (const.s) t' c Hconst') :=
          have P : t = typ.arr typ.nat typ.nat, from const.has_typ.s.inv Hconst,
          have P' : t' = typ.arr typ.nat typ.nat, from const.has_typ.s.inv Hconst',
          eq.subst (eq.symm P') P
      | inj (@has_typ.lam g g' x t t' e Hctx Htyp) (@has_typ.lam g _g' x t _t' e _Hctx _Htyp) :=
          have _Geq : _g' = g', from table.is_unionr.inj _Hctx Hctx,
          have _Teq : _t' = t', from inj (eq.subst _Geq _Htyp) Htyp,
          eq.subst _Teq rfl
      | inj (@has_typ.ap c _t t e1 e2 H1 H2) (@has_typ.ap c _t' t' e1 e2 H1' H2') :=
          have P : typ.arr _t t = typ.arr _t' t', from inj H1 H1',
          typ.no_confusion P (fun _eq eq, eq)
    end has_typ
  end exp
end lnat
